{-

This module defines the Graph data type. The fields are,
1) nodes:    set of all the nodes in this graph
2) inEdges:  map from a node to a list of nodes from which it is linked
3) outEdges: map from a node to a list of nodes to which it links

This module also contains some utility functions for this data type.

-}

module Graph
( Graph(..)
, Node
, fromFile
) where

import qualified Data.Map as M (Map, insertWith, empty, keysSet)
import qualified Data.Set as S (Set, union, fromList, empty, toList, difference)
import System.IO (readFile)

type Node = String
type Nodes = S.Set Node
type InEdges = M.Map Node [Node]
type OutEdges = M.Map Node [Node]

data Graph = Graph { nodes    :: Nodes
                   , inEdges  :: InEdges
                   , outEdges :: OutEdges } deriving Show

-- Initial state of an empty graph                   
empty :: Graph
empty = Graph S.empty M.empty M.empty

-- Read in a graph from a file
fromFile :: String -> IO Graph
fromFile filename = do 
    content <- readFile filename
    return $ fromContent content

fromContent :: String -> Graph
fromContent content =
    let ls = lines content
    in postProcess $ foldl parseLine empty ls
    where 
        postProcess :: Graph -> Graph
        postProcess graph = foldl parseLine graph newLines
            where 
                ns = nodes graph
                sinkNodes = S.difference ns $ M.keysSet $ outEdges graph
                newLines = [ n1 ++ " " ++ n2 |
                 n1 <- S.toList sinkNodes, n2 <- S.toList ns, n1 /= n2 ]

{-
  Parse each line of the input file as an edge in the graph.
  Each line of the input file should conform to the format `fromNode toNode`.
-}
parseLine :: Graph -> String -> Graph
parseLine graph line =
    let ws = words line
    in case ws of
        [fromNode, toNode] ->
            Graph ns iEdges oEdges
            where 
                ns = S.union (S.fromList ws) (nodes graph)
                iEdges = M.insertWith (++) toNode [fromNode] (inEdges graph)
                oEdges = M.insertWith (++) fromNode [toNode] (outEdges graph)
        _ -> error "All lines of the input file \
        \should be in the format of `fromNode toNode`"
