{-

This module defines the PageRank data type, which is a map from a node to its
current page rank value.

This module also contains the empty definition and some utility functions for
this data type, such as the mapper and the reducer functions to compute the page
rank values for a given graph with MapReduce, and also a sequential method to
compute page rank.

-}

module PageRank
( PageRank
, initFromGraph
, computePageRankSeq
, computePageRankMR
) where

import Graph (Graph(..), Node)
import MapReduce (mapReduce)
import qualified Data.Map as M (Map, empty, fromList, lookup, unionWith, toList)
import qualified Data.Set as S (toList, size)
import Data.Maybe (fromJust)

type PageRankValue = Double
type PageRank = M.Map Node PageRankValue

-- Initial state of a PageRank data for an empty graph
empty :: PageRank
empty = M.empty

{-
  Initial state of a PageRank data for a given graph, the page rank value
  of each node is the reciprocal of the number of nodes in this graph
-}
initFromGraph :: Graph -> PageRank
initFromGraph graph =
    let ns = nodes graph
        pr = 1.0 / (fromIntegral $ S.size ns) in
    M.fromList [ (n, pr) | n <- S.toList ns ]

mapper :: (PageRankValue, [Node]) -> PageRank
mapper (pr, outNodes) = 
    let pr_ = pr / (fromIntegral $ length outNodes) in 
    M.fromList [ (n, pr_) | n <- outNodes ]

reducer :: [PageRank] -> PageRank
reducer [] = empty
reducer [x] = x
reducer (x:xs) = M.unionWith (+) x (reducer xs)

{-
  Given initial PageRank and the corresponding graph, a number of iterations
  to compute, and a damping factor, returns the resulting PageRank after those
  iterations of computation in a parallel way with MapReduce
-}
computePageRankMR :: PageRank -> Graph -> Int -> Double -> PageRank
computePageRankMR pageRank _ 0 _ = pageRank
computePageRankMR pageRank graph itrs damping =
    let nextPageRank = computeNextPageRankMR pageRank
    in computePageRankMR nextPageRank graph (itrs-1) damping
    where
        computeNextPageRankMR :: PageRank -> PageRank
        computeNextPageRankMR curPR = 
            let ns = S.toList $ nodes graph
                input = map produceInput ns
                produceInput n = (pr, outNodes)
                    where
                        pr = fromJust $ M.lookup n curPR
                        outNodes = fromJust $ M.lookup n $ outEdges graph
                mrResult = mapReduce mapper reducer input
                base = (1 - damping) / (fromIntegral $ length ns)
            in M.fromList [ (n, base + damping * pr) | (n, pr) <- M.toList mrResult ]

{-
  Given initial PageRank and the corresponding graph, a number of iterations
  to compute, and a damping factor, returns the resulting PageRank after those
  iterations of computation in a sequential way
-}
computePageRankSeq :: PageRank -> Graph -> Int -> Double -> PageRank
computePageRankSeq pageRank _ 0 _ = pageRank
computePageRankSeq pageRank graph itrs damping =
    let nextPageRank = computeNextPageRank pageRank
    in computePageRankSeq nextPageRank graph (itrs-1) damping
    where
        computeNextPageRank :: PageRank -> PageRank
        computeNextPageRank curPR =
            M.fromList [ (n, computePRValue n) | n <- ns ]
            where
                ns = S.toList $ nodes graph
                iEdges = inEdges graph
                oEdges = outEdges graph
                computePRValue :: Node -> PageRankValue
                computePRValue n =
                    let inNodes = fromJust $ M.lookup n iEdges 
                    in (1 - damping) / (fromIntegral $ length ns) + damping * (foldl sumUp 0 inNodes)
                    where 
                        sumUp acc node = 
                            let numOutNodes = length $ fromJust $ M.lookup node oEdges
                                prValue = fromJust $ M.lookup node curPR
                            in acc + prValue / (fromIntegral numOutNodes)
