{-
The benchmark PageRank computation implementation based on external opensourced MapReduce library
GitHub repository of the MapReduce library: https://github.com/jdstmporter/Haskell-MapReduce
-}

module PageRankExt
(
computePageRankMRext
) where

import Graph (Graph(..), Node)
import qualified Data.Map as M (Map, empty, fromList, lookup, unionWith, toList)
import qualified Data.Set as S (toList, size)
import Data.Maybe (fromJust)
import MapReduceLibExt (run,distribute,lift,(>>=))

type PageRankValue = Double
type PageRank = M.Map Node PageRankValue

empty :: PageRank
empty = M.empty

initFromGraph :: Graph -> PageRank
initFromGraph graph =
    let ns = nodes graph
        pr = 1.0 / (fromIntegral $ S.size ns) in
    M.fromList [ (n, pr) | n <- S.toList ns ]

mr :: Double -> Double -> Int -> [(Node, (PageRankValue, [Node]))] -> [(Node, (PageRankValue, [Node]))]
mr damping numNodes n state = run f state
    where
        f = distribute n MapReduceLibExt.>>= lift mapper MapReduceLibExt.>>= lift (reducer damping numNodes)

-- According to the specification of Haskell-MapReduce lib
-- mapper should take the form of [s] -> [(s', a)]
-- where s is input data, s' is output data and a is output key
mapper :: [(Node, (PageRankValue, [Node]))] -> [((Node, (PageRankValue, [Node])), Node)]
mapper [] = []
mapper (x:xs) = parse x ++ mapper xs
    where
        parse (n, (pr, outNodes)) =
            let pr_ = pr / (fromIntegral $ length outNodes)
            in ((n, (0, outNodes)), n) : [ ((n_, (pr_, [])), n_) | n_ <- outNodes ]

-- According to the specification of Haskell-MapReduce lib
-- reducer should take the form of [s'] -> [s'']
-- where s' is output data of mapper, s'' is output data of reducer
reducer :: Double -> Double -> [(Node, (PageRankValue, [Node]))] -> [(Node, (PageRankValue, [Node]))]
reducer _ _ [] = []
reducer damping numNodes xs@(x:_) =
    [ foldl f (fst x, ((1 - damping) / numNodes, [])) xs ]
    where f x y = (
            fst x, 
            (
                (fst $ snd x) + damping * (fst $ snd y), 
                (snd $ snd x) ++ (snd $ snd y)
            )
            )

computePageRankMRext :: PageRank -> Graph -> Int -> Double -> PageRank
computePageRankMRext pageRank _ 0 _ = pageRank
computePageRankMRext pageRank graph itrs damping =
    let ns = S.toList $ nodes graph
        numNodes = fromIntegral $ length ns
        oEdges = outEdges graph
        initMRinput = map toMRinput ns
        toMRinput n =
            let pr = fromJust $ M.lookup n pageRank
                outNodes = fromJust $ M.lookup n oEdges
            in (n, (pr, outNodes))
        mrOutput = mrItr initMRinput damping numNodes itrs
    in M.fromList [ (n, pr) | (n, (pr, _)) <- mrOutput ]
    where
        mrItr :: [(Node, (PageRankValue, [Node]))] -> Double -> Double -> Int -> [(Node, (PageRankValue, [Node]))]
        mrItr input _ _ 0 = input
        mrItr input damping numNodes itrs =
            let output = mr damping numNodes 6 input
            in mrItr output damping numNodes (itrs-1)
